@@ -5,7 +5,7 @@ from typing import (
 
 # We support these resource types. The order matters because it determines the
 # order in which the manifests will be grouped in the output files.
-SUPPORTED_KINDS = (
+DEFAULT_PRIORITIES = (
     # Namespaces must come first to ensure the other resources can be created
     # within them.
     "Namespace",
@@ -29,42 +29,6 @@ SUPPORTED_KINDS = (
     "Ingress",
 )
 
-
-# These resource exist outside of namespaces.
-NON_NAMESPACED_KINDS = {
-    "Namespace", "ClusterRole", "ClusterRoleBinding", "CustomResourceDefinition"
-}
-
-# Declare aliases for each resource type. Will be used in command line parsing
-# to save the user some typing and match what `kubectl` would accept. We do not
-# need to worry about capitalisation because `square.parse_commandline_args`
-# will always convert everything to lower case letters first.
-RESOURCE_ALIASES = {
-    "ClusterRole": {"clusterrole", "clusterroles"},
-    "ClusterRoleBinding": {"clusterrolebinding", "clusterrolebindings"},
-    "ConfigMap": {"configmap", "cm"},
-    "CronJob": {"cronjob", "cj"},
-    "CustomResourceDefinition": {"crd", "customresourcedefinitions"},
-    "DaemonSet": {"daemonset", "daemonsets", "ds"},
-    "Deployment": {"deployment", "deployments", "deploy"},
-    "HorizontalPodAutoscaler": {"hpa"},
-    "Ingress": {"ingress", "ingresses", "ing"},
-    "Namespace": {"namespace", "namespaces", "ns"},
-    "PersistentVolumeClaim": {"persistentVolumeClaim", "persistentvolumeclaims", "pvc"},
-    "PodDisruptionBudget": {"poddisruptionbudget", "pdb"},
-    "Role": {"role", "roles"},
-    "RoleBinding": {"rolebinding", "rolebindings"},
-    "Secret": {"secret", "secrets"},
-    "Service": {"service", "services", "svc"},
-    "ServiceAccount": {"serviceaccount", "serviceaccounts", "sa"},
-    "StatefulSet": {"statefulset", "statefulsets"},
-}
-
-# Sanity check: we must have aliases for every supported resource kind.
-assert set(SUPPORTED_KINDS) == set(RESOURCE_ALIASES.keys())
-
-SUPPORTED_VERSIONS = ("1.9", "1.10", "1.11", "1.12", "1.13", "1.14", "1.15")
-
 Filepath = pathlib.Path
 
 
@@ -76,6 +40,28 @@ class K8sClientCert(NamedTuple):
     key: Filepath = Filepath()
 
 
+class MetaManifest(NamedTuple):
+    """Minimum amount of information to uniquely identify a K8s resource.
+
+    The primary purpose of this tuple is as to provide an immutable UUID that
+    we can use as keys in dictionaries or entries in a set.
+
+    """
+    apiVersion: str
+    kind: str
+    namespace: Optional[str]
+    name: str
+
+
+class K8sResource(NamedTuple):
+    """Describe a specific K8s resource kind."""
+    apiVersion: str   # "batch/v1beta1" or "extensions/v1beta1".
+    kind: str         # "Deployment" (as specified in manifest)
+    name: str         # "deployment" (usually lower case version of above)
+    namespaced: bool  # Whether or not the resource is namespaced.
+    url: str          # API endpoint, eg "k8s-host.com//api/v1/pods".
+
+
 class K8sConfig(NamedTuple):
     """Everything we need to know to connect and authenticate with Kubernetes."""
     url: str = ""               # Kubernetes API
@@ -86,17 +72,24 @@ class K8sConfig(NamedTuple):
     ca_cert: Filepath = Filepath()
     client_cert: Optional[K8sClientCert] = None
 
+    # Request session.
+    client: Any = None
+
     # Kubernetes version and name.
     version: str = ""
     name: str = ""
 
+    # Kubernetes API endpoints (see k8s.compile_api_endpoints).
+    apis: Dict[Tuple[str, str], K8sResource] = {}
 
-class MetaManifest(NamedTuple):
-    """A succinct summary of a K8s resource manifest."""
-    apiVersion: str
-    kind: str
-    namespace: str
-    name: str
+    # LUT to translate short names into their proper resource kind,
+    # for instance short = {"service":, "Service", "svc": "Service"}
+    short2kind: Dict[str, str] = {}
+
+    # The set of supported K8s resource kinds, eg {"Deployment", "Service"}.
+    # NOTE: these are the `manifest.kind` spellings. "Deployment" is a kind
+    # whereas "deployment" or "Deployments" are not.
+    kinds: Set[str] = set()
 
 
 # -----------------------------------------------------------------------------
@@ -153,7 +146,7 @@ class DeploymentPlanMeta(NamedTuple):
 # -----------------------------------------------------------------------------
 class Selectors(NamedTuple):
     """Comprises all the filters to select manifests."""
-    kinds: Iterable[str]
+    kinds: Set[str]
     namespaces: Optional[Iterable[str]]
     labels: Optional[Set[Tuple[str, str]]]
 
@@ -164,21 +157,27 @@ class GroupBy(NamedTuple):
     order: Iterable[str]        # ["ns", "label=app", kind"]
 
 
-class Configuration(NamedTuple):
-    """Square configuration.
-
-    This will be compiled at startup but is immutable afterwards.
-
-    """
-    command: str
+class Config(NamedTuple):
+    """Uniform interface into top level Square API."""
+    # Path to local manifests eg "./foo"
     folder: Filepath
-    selectors: Selectors
-    groupby: GroupBy
+
+    # Path to Kubernetes credentials.
     kubeconfig: Filepath
-    kube_ctx: Optional[str] = None
-    verbosity: int = 0
-    k8s_config: K8sConfig = K8sConfig()
-    k8s_client: Any = None
+
+    # Kubernetes context (use `None` to use the default).
+    kube_ctx: Optional[str]
+
+    # Only operate on resources that match the selectors.
+    selectors: Selectors = Selectors(kinds=set(DEFAULT_PRIORITIES),
+                                     namespaces=None,
+                                     labels=set())
+
+    # Sort the manifest in this order, or alphabetically at the end if not in the list.
+    priorities: Collection[str] = tuple(DEFAULT_PRIORITIES)
+
+    # How to structure the folder directory when syncing manifests.
+    groupby: GroupBy = GroupBy("", tuple())
 
 
 # -----------------------------------------------------------------------------
