@@ -7,14 +7,15 @@ import re
 import subprocess
 import tempfile
 import warnings
-from typing import Optional, Tuple
+from collections import defaultdict
+from typing import Dict, List, Optional, Set, Tuple
 
 import google.auth
 import google.auth.transport.requests
 import requests
 import yaml
 from square.dtypes import (
-    SUPPORTED_KINDS, SUPPORTED_VERSIONS, Filepath, K8sClientCert, K8sConfig,
+    Filepath, K8sClientCert, K8sConfig, K8sResource, MetaManifest,
 )
 
 FNAME_TOKEN = Filepath("/var/run/secrets/kubernetes.io/serviceaccount/token")
@@ -25,10 +26,8 @@ FNAME_CERT = Filepath("/var/run/secrets/kubernetes.io/serviceaccount/ca.crt")
 logit = logging.getLogger("square")
 
 
-def load_kubeconfig(
-        fname: Filepath,
-        context: Optional[str]
-) -> Tuple[Optional[str], Optional[dict], Optional[dict]]:
+def load_kubeconfig(fname: Filepath,
+                    context: Optional[str]) -> Tuple[str, dict, dict, bool]:
     """Return user name and user- and cluster information.
 
     Return None on error.
@@ -48,7 +47,7 @@ def load_kubeconfig(
         kubeconf = yaml.safe_load(open(fname))
     except (IOError, PermissionError) as err:
         logit.error(f"{err}")
-        return (None, None, None)
+        return ("", {}, {}, True)
 
     # Find the user and cluster information based on the specified `context`.
     try:
@@ -71,7 +70,7 @@ def load_kubeconfig(
             assert len(user_info) == len(cluster_info) == 1
         except AssertionError:
             logit.error(f"Could not find information for context <{ctx}>")
-            return (None, None, None)
+            return ("", {}, {}, True)
 
         # Unpack the cluster- and user information.
         cluster_name = cluster_info[0]["name"]
@@ -81,16 +80,16 @@ def load_kubeconfig(
         del cluster_info
     except (KeyError, TypeError):
         logit.error(f"Kubeconfig YAML file <{fname}> is invalid")
-        return (None, None, None)
+        return ("", {}, {}, True)
 
     # Success. The explicit `dicts()` are to satisfy MyPy.
     logit.info(f"Loaded {ctx} from Kubeconfig file <{fname}>")
-    return (username, dict(user_info), dict(cluster_info_out))
+    return (username, dict(user_info), dict(cluster_info_out), False)
 
 
 def load_incluster_config(
         fname_token: Filepath = FNAME_TOKEN,
-        fname_cert: Filepath = FNAME_CERT) -> Optional[K8sConfig]:
+        fname_cert: Filepath = FNAME_CERT) -> Tuple[K8sConfig, bool]:
     """Return K8s access config from Pod service account.
 
     Returns None if we are not running in a Pod.
@@ -116,7 +115,7 @@ def load_incluster_config(
         assert fname_token.exists()
     except AssertionError:
         logit.debug("Could not find incluster (service account) credentials.")
-        return None
+        return K8sConfig(), True
 
     # Return the compiled K8s access configuration.
     logit.info("Use incluster (service account) credentials.")
@@ -127,13 +126,13 @@ def load_incluster_config(
         client_cert=None,
         version="",
         name="",
-    )
+    ), False
 
 
 def load_gke_config(
         fname: Filepath,
         context: Optional[str],
-        disable_warnings: bool = False) -> Optional[K8sConfig]:
+        disable_warnings: bool = False) -> Tuple[K8sConfig, bool]:
     """Return K8s access config for GKE cluster described in `kubeconfig`.
 
     Returns None if `kubeconfig` does not exist or could not be parsed.
@@ -151,9 +150,9 @@ def load_gke_config(
 
     """
     # Parse the kubeconfig file.
-    name, user, cluster = load_kubeconfig(fname, context)
-    if name is None or user is None or cluster is None:
-        return None
+    name, user, cluster, err = load_kubeconfig(fname, context)
+    if err:
+        return (K8sConfig(), True)
 
     # Unpack the self signed certificate (Google does not register the K8s API
     # server certificate with a public CA).
@@ -161,7 +160,7 @@ def load_gke_config(
         ssl_ca_cert_data = base64.b64decode(cluster["certificate-authority-data"])
     except KeyError:
         logit.debug(f"Context {context} in <{fname}> is not a GKE config")
-        return None
+        return (K8sConfig(), True)
 
     # Save the certificate to a temporary file. This is only necessary because
     # the requests library will need a path to the CA file - unfortunately, we
@@ -186,13 +185,13 @@ def load_gke_config(
         client_cert=None,
         version="",
         name=cluster["name"],
-    )
+    ), False
 
 
 def load_eks_config(
         fname: Filepath,
         context: Optional[str],
-        disable_warnings: bool = False) -> Optional[K8sConfig]:
+        disable_warnings: bool = False) -> Tuple[K8sConfig, bool]:
     """Return K8s access config for EKS cluster described in `kubeconfig`.
 
     Returns None if `kubeconfig` does not exist or could not be parsed.
@@ -210,9 +209,9 @@ def load_eks_config(
 
     """
     # Parse the kubeconfig file.
-    name, user, cluster = load_kubeconfig(fname, context)
-    if name is None or user is None or cluster is None:
-        return None
+    name, user, cluster, err = load_kubeconfig(fname, context)
+    if err:
+        return (K8sConfig(), True)
 
     # Get a copy of all env vars. We will pass that one along to the
     # sub-process, plus the env vars specified in the kubeconfig file.
@@ -227,7 +226,7 @@ def load_eks_config(
         env_kubeconf = user["exec"].get("env", [])
     except KeyError:
         logit.debug(f"Context {context} in <{fname}> is not an EKS config")
-        return None
+        return (K8sConfig(), True)
 
     # Convert a None value (valid value in YAML) to an empty list of env vars.
     env_kubeconf = env_kubeconf if env_kubeconf is not None else []
@@ -252,10 +251,10 @@ def load_eks_config(
         token = yaml.safe_load(out.stdout.decode("utf8"))["status"]["token"]
     except FileNotFoundError:
         logit.error(f"Could not find <{cmd}> application to get token")
-        return None
+        return (K8sConfig(), True)
     except (KeyError, yaml.YAMLError):
         logit.error(f"Token manifest from <{cmd}> is corrupt")
-        return None
+        return (K8sConfig(), True)
 
     # Return the config data.
     logit.info(f"Assuming EKS cluster.")
@@ -266,13 +265,11 @@ def load_eks_config(
         client_cert=None,
         version="",
         name=cluster["name"],
-    )
+    ), False
 
 
-def load_minikube_config(
-        fname: Filepath,
-        context: Optional[str],
-) -> Optional[K8sConfig]:
+def load_minikube_config(fname: Filepath,
+                         context: Optional[str]) -> Tuple[K8sConfig, bool]:
     """Load minikube configuration from `fname`.
 
     Return None on error.
@@ -288,9 +285,9 @@ def load_minikube_config(
 
     """
     # Parse the kubeconfig file.
-    name, user, cluster = load_kubeconfig(fname, context)
-    if name is None or user is None or cluster is None:
-        return None
+    name, user, cluster, err = load_kubeconfig(fname, context)
+    if err:
+        return (K8sConfig(), True)
 
     # Minikube uses client certificates to authenticate. We need to pass those
     # to the HTTP client of our choice when we create the session.
@@ -309,16 +306,13 @@ def load_minikube_config(
             client_cert=client_cert,
             version="",
             name=cluster["name"],
-        )
+        ), False
     except KeyError:
         logit.debug(f"Context {context} in <{fname}> is not a Minikube config")
-        return None
+        return (K8sConfig(), True)
 
 
-def load_kind_config(
-        fname: Filepath,
-        context: Optional[str],
-) -> Optional[K8sConfig]:
+def load_kind_config(fname: Filepath, context: Optional[str]) -> Tuple[K8sConfig, bool]:
     """Load Kind configuration from `fname`.
 
     https://github.com/bsycorp/kind
@@ -340,9 +334,9 @@ def load_kind_config(
 
     """
     # Parse the kubeconfig file.
-    name, user, cluster = load_kubeconfig(fname, context)
-    if name is None or user is None or cluster is None:
-        return None
+    name, user, cluster, err = load_kubeconfig(fname, context)
+    if err:
+        return (K8sConfig(), True)
 
     # Kind/Minikube use client certificates to authenticate. We need to pass
     # those to the HTTP client of our choice when we create the session.
@@ -368,16 +362,16 @@ def load_kind_config(
             client_cert=client_cert,
             version="",
             name=cluster["name"],
-        )
+        ), False
     except KeyError:
         logit.debug(f"Context {context} in <{fname}> is not a Minikube config")
-        return None
+        return (K8sConfig(), True)
 
 
 def load_auto_config(
         fname: Filepath,
         context: Optional[str],
-        disable_warnings: bool = False) -> Optional[K8sConfig]:
+        disable_warnings: bool = False) -> Tuple[K8sConfig, bool]:
     """Automagically find and load the correct K8s configuration.
 
     This function will load several possible configuration options and returns
@@ -398,155 +392,136 @@ def load_auto_config(
         Config
 
     """
-    conf = load_incluster_config()
-    if conf is not None:
-        return conf
+    conf, err = load_incluster_config()
+    if not err:
+        return conf, False
     logit.debug("Incluster config failed")
 
-    conf = load_minikube_config(fname, context)
-    if conf is not None:
-        return conf
+    conf, err = load_minikube_config(fname, context)
+    if not err:
+        return conf, False
     logit.debug("Minikube config failed")
 
-    conf = load_kind_config(fname, context)
-    if conf is not None:
-        return conf
+    conf, err = load_kind_config(fname, context)
+    if not err:
+        return conf, False
     logit.debug("KIND config failed")
 
-    conf = load_eks_config(fname, context, disable_warnings)
-    if conf is not None:
-        return conf
+    conf, err = load_eks_config(fname, context, disable_warnings)
+    if not err:
+        return conf, False
     logit.debug("EKS config failed")
 
-    conf = load_gke_config(fname, context, disable_warnings)
-    if conf is not None:
-        return conf
+    conf, err = load_gke_config(fname, context, disable_warnings)
+    if not err:
+        return conf, False
     logit.debug("GKE config failed")
 
     logit.error(f"Could not find a valid configuration in <{fname}>")
-    return None
+    return (K8sConfig(), True)
 
 
-def session(config: K8sConfig):
+def session(k8sconfig: K8sConfig):
     """Return configured `requests` session."""
     # Plain session.
     sess = requests.Session()
 
     # Load the CA file (necessary for self signed certs to avoid https warning).
-    sess.verify = str(config.ca_cert)
+    sess.verify = str(k8sconfig.ca_cert)
 
     # Add the client certificate, if the cluster uses those to authenticate users.
-    if config.client_cert is not None:
-        sess.cert = (str(config.client_cert.crt), str(config.client_cert.key))
+    if k8sconfig.client_cert is not None:
+        sess.cert = (str(k8sconfig.client_cert.crt), str(k8sconfig.client_cert.key))
 
     # Add the bearer token if this cluster uses them to authenticate users.
-    if config.token is not None:
-        sess.headers.update({'authorization': f'Bearer {config.token}'})
+    if k8sconfig.token is not None:
+        sess.headers.update({'authorization': f'Bearer {k8sconfig.token}'})
 
     # Return the configured session object.
     return sess
 
 
-def urlpath(
-        config: K8sConfig,
-        kind: str,
-        namespace: Optional[str]) -> Tuple[Optional[str], bool]:
-    """Return complete URL to K8s resource.
+def resource(k8sconfig: K8sConfig, meta: MetaManifest) -> Tuple[K8sResource, bool]:
+    """Return `K8sResource` object.
+
+    That object will contain the full path to a resource, eg.
+    https://1.2.3.4/api/v1/namespace/foo/services.
 
     Inputs:
-        config: k8s.Config
-        kind: str
-            Eg "Deployment", "Namespace", ... (case sensitive).
-        namespace: str
-            Must be None for "Namespace" resources.
+        k8sconfig: K8sConfig
+        meta: MetaManifest
 
     Returns:
-        str: Full path K8s resource, eg https://1.2.3.4/api/v1/namespace/foo/services.
+        K8sResource
 
     """
+    err_resp = (K8sResource("", "", "", False, ""), True)
+
+    # Compile the lookup key for the resource, eg `("Service", "v1")`.
+    if not meta.apiVersion:
+        # Use the most recent version of the API if None was specified.
+        candidates = [(kind, ver) for kind, ver in k8sconfig.apis if kind == meta.kind]
+        if len(candidates) == 0:
+            logit.warning(f"Cannot determine API version for <{meta.kind}>")
+            return err_resp
+        candidates.sort()
+        key = candidates.pop(0)
+    else:
+        key = (meta.kind, meta.apiVersion)
+
+    # Retrieve the resource.
+    try:
+        resource = k8sconfig.apis[key]
+    except KeyError:
+        logit.error(f"Unsupported resource <{meta.kind}> {key}.")
+        return err_resp
+
+    # Void the "namespace" key for non-namespaced resources.
+    if not resource.namespaced:
+        meta = meta._replace(namespace=None)
+
     # Namespaces are special because they lack the `namespaces/` path prefix.
-    if kind == "Namespace" or namespace is None:
+    if meta.kind == "Namespace":
+        # Return the correct URL, depending on whether we want all namespaces
+        # or a particular one.
+        url = f"{resource.url}/namespaces"
+        if meta.name:
+            url += f"/{meta.name}"
+        return resource._replace(url=url), False
+
+    # Determine if the prefix for namespaced resources.
+    if meta.namespace is None:
         namespace = ""
     else:
         # Namespace name must conform to K8s standards.
-        match = re.match(r"[a-z0-9]([-a-z0-9]*[a-z0-9])?", namespace)
-        if match is None or match.group() != namespace:
-            logit.error(f"Invalid namespace name <{namespace}>.")
-            return (None, True)
+        match = re.match(r"[a-z0-9]([-a-z0-9]*[a-z0-9])?", meta.namespace)
+        if match is None or match.group() != meta.namespace:
+            logit.error(f"Invalid namespace name <{meta.namespace}>.")
+            return err_resp
+        namespace = f"namespaces/{meta.namespace}"
 
-        namespace = f"namespaces/{namespace}/"
+    # Sanity check: we cannot search for a namespaced resource by name in all
+    # namespaces. Example: we cannot search for a Service `foo` in all
+    # namespaces. We could only search for Service `foo` in namespace `bar`, or
+    # all services in all namespaces.
+    if resource.namespaced and meta.name and not meta.namespace:
+        logit.error(f"Cannot search for {meta.kind} {meta.name} in {meta.namespace}")
+        return err_resp
 
-    # We must support the specified resource kind.
-    if kind not in SUPPORTED_KINDS:
-        logit.error(f"Unsupported resource <{kind}>.")
-        return (None, True)
+    # Create the full path to the resource depending on whether we have a
+    # namespace and resource name. Here are all three possibilities:
+    #  - /api/v1/namespaces/services
+    #  - /api/v1/namespaces/my-namespace/services
+    #  - /api/v1/namespaces/my-namespace/services/my-service
+    path = f"{namespace}/{resource.name}" if namespace else resource.name
+    path = f"{path}/{meta.name}" if meta.name else path
 
-    # We must support the specified K8s version.
-    if config.version not in SUPPORTED_VERSIONS:
-        logit.error(f"Unsupported K8s version <{config.version}>.")
-        return (None, True)
-
-    # The HTTP request path names by K8s version and resource kind.
-    # The keys in this dict must cover all those specified in
-    # `SUPPORTED_VERSIONS` and `SUPPORTED_KINDS`.
-    resources = {
-        "1.9": {
-            "ClusterRole": f"apis/rbac.authorization.k8s.io/v1/clusterroles",
-            "ClusterRoleBinding": f"apis/rbac.authorization.k8s.io/v1/clusterrolebindings",  # NOQA
-            "ConfigMap": f"api/v1/{namespace}/configmaps",
-            "CronJob": f"apis/batch/v1beta1/{namespace}/cronjobs",
-            "CustomResourceDefinition": f"apis/apiextensions.k8s.io/v1beta1/customresourcedefinitions",  # NOQA
-            "DaemonSet": f"apis/apps/v1/{namespace}/daemonsets",
-            "Deployment": f"apis/extensions/v1beta1/{namespace}/deployments",
-            "HorizontalPodAutoscaler": f"apis/autoscaling/v1/{namespace}/horizontalpodautoscalers",   # NOQA
-            "Ingress": f"apis/extensions/v1beta1/{namespace}/ingresses",
-            "Namespace": f"api/v1/namespaces",
-            "PersistentVolumeClaim": f"api/v1/{namespace}/persistentvolumeclaims",
-            "PodDisruptionBudget": f"apis/policy/v1beta1/{namespace}/poddisruptionbudgets",  # NOQA
-            "Role": f"apis/rbac.authorization.k8s.io/v1/{namespace}/roles",
-            "RoleBinding": f"apis/rbac.authorization.k8s.io/v1/{namespace}/rolebindings",
-            "Secret": f"api/v1/{namespace}/secrets",
-            "Service": f"api/v1/{namespace}/services",
-            "ServiceAccount": f"api/v1/{namespace}/serviceaccounts",
-            "StatefulSet": f"apis/apps/v1/{namespace}/statefulsets",
-        },
-        "1.11": {
-            "ClusterRole": f"apis/rbac.authorization.k8s.io/v1/clusterroles",
-            "ClusterRoleBinding": f"apis/rbac.authorization.k8s.io/v1/clusterrolebindings",  # NOQA
-            "ConfigMap": f"api/v1/{namespace}/configmaps",
-            "CronJob": f"apis/batch/v1beta1/{namespace}/cronjobs",
-            "CustomResourceDefinition": f"apis/apiextensions.k8s.io/v1beta1/customresourcedefinitions",  # NOQA
-            "DaemonSet": f"apis/apps/v1/{namespace}/daemonsets",
-            "Deployment": f"apis/apps/v1/{namespace}/deployments",
-            "HorizontalPodAutoscaler": f"apis/autoscaling/v1/{namespace}/horizontalpodautoscalers",   # NOQA
-            "Ingress": f"apis/extensions/v1beta1/{namespace}/ingresses",
-            "Namespace": f"api/v1/namespaces",
-            "PersistentVolumeClaim": f"api/v1/{namespace}/persistentvolumeclaims",
-            "PodDisruptionBudget": f"apis/policy/v1beta1/{namespace}/poddisruptionbudgets",  # NOQA
-            "Role": f"apis/rbac.authorization.k8s.io/v1/{namespace}/roles",
-            "RoleBinding": f"apis/rbac.authorization.k8s.io/v1/{namespace}/rolebindings",
-            "Secret": f"api/v1/{namespace}/secrets",
-            "Service": f"api/v1/{namespace}/services",
-            "ServiceAccount": f"api/v1/{namespace}/serviceaccounts",
-            "StatefulSet": f"apis/apps/v1/{namespace}/statefulsets",
-        },
-    }
-
-    # Replicate the resource URLs for those clusters where they match.
-    resources["1.10"] = resources["1.9"]
-    resources["1.12"] = resources["1.11"]
-    resources["1.13"] = resources["1.11"]
-    resources["1.14"] = resources["1.11"]
-    resources["1.15"] = resources["1.11"]
-
-    # Look up the resource path and remove duplicate "/" characters (may have
-    # slipped in when no namespace was supplied, eg "/api/v1//configmaps").
-    path = resources[config.version][kind]
+    # The concatenation above may have introduced `//`. Here we remove them.
     path = path.replace("//", "/")
-    assert not path.startswith("/"), path
 
-    # Return the complete resource URL.
-    return (f"{config.url}/{path}", False)
+    # Return the K8sResource with the correct URL.
+    resource = resource._replace(url=f"{resource.url}/{path}")
+    return resource, False
 
 
 def request(
@@ -554,7 +529,7 @@ def request(
         method: str,
         url: str,
         payload: Optional[dict],
-        headers: Optional[dict]) -> Tuple[Optional[dict], bool]:
+        headers: Optional[dict]) -> Tuple[dict, bool]:
     """Return response of web request made with `client`.
 
     Inputs:
@@ -577,7 +552,7 @@ def request(
         method = err.request.method
         url = err.request.url
         logit.error(f"Connection error: {method} {url}")
-        return (None, True)
+        return ({}, True)
 
     try:
         response = ret.json()
@@ -587,7 +562,7 @@ def request(
             "-" * 80 + "\n" + err.doc + "\n" + "-" * 80,
         )
         logit.error(str.join("\n", msg))
-        return (None, True)
+        return ({}, True)
 
     # Log the entire request in debug mode.
     logit.debug(
@@ -599,7 +574,7 @@ def request(
     return (response, ret.status_code)
 
 
-def delete(client, url: str, payload: dict) -> Tuple[Optional[dict], bool]:
+def delete(client, url: str, payload: dict) -> Tuple[dict, bool]:
     """Make DELETE requests to K8s (see `k8s_request`)."""
     resp, code = request(client, 'DELETE', url, payload, headers=None)
     err = (code not in (200, 202))
@@ -608,7 +583,7 @@ def delete(client, url: str, payload: dict) -> Tuple[Optional[dict], bool]:
     return (resp, err)
 
 
-def get(client, url: str) -> Tuple[Optional[dict], bool]:
+def get(client, url: str) -> Tuple[dict, bool]:
     """Make GET requests to K8s (see `request`)."""
     resp, code = request(client, 'GET', url, payload=None, headers=None)
     err = (code != 200)
@@ -617,7 +592,7 @@ def get(client, url: str) -> Tuple[Optional[dict], bool]:
     return (resp, err)
 
 
-def patch(client, url: str, payload: dict) -> Tuple[Optional[dict], bool]:
+def patch(client, url: str, payload: dict) -> Tuple[dict, bool]:
     """Make PATCH requests to K8s (see `request`)."""
     headers = {'Content-Type': 'application/json-patch+json'}
     resp, code = request(client, 'PATCH', url, payload, headers)
@@ -627,7 +602,7 @@ def patch(client, url: str, payload: dict) -> Tuple[Optional[dict], bool]:
     return (resp, err)
 
 
-def post(client, url: str, payload: dict) -> Tuple[Optional[dict], bool]:
+def post(client, url: str, payload: dict) -> Tuple[dict, bool]:
     """Make POST requests to K8s (see `request`)."""
     resp, code = request(client, 'POST', url, payload, headers=None)
     err = (code != 201)
@@ -636,26 +611,25 @@ def post(client, url: str, payload: dict) -> Tuple[Optional[dict], bool]:
     return (resp, err)
 
 
-def version(config: K8sConfig, client) -> Tuple[Optional[K8sConfig], bool]:
-    """Return new `config` with version number of K8s API.
+def version(k8sconfig: K8sConfig) -> Tuple[K8sConfig, bool]:
+    """Return new `k8sconfig` with version number of K8s API.
 
-    Contact the K8s API, query its version via `client` and return `config`
-    with an updated `version` field. All other field in `config` will remain
+    Contact the K8s API, query its version via `client` and return `k8sconfig`
+    with an updated `version` field. All other field in `k8sconfig` will remain
     intact.
 
     Inputs:
-        config: Config
-        client: `requests` session with correct K8s certificates.
+        k8sconfig: K8sConfig
 
     Returns:
-        Config
+        K8sConfig
 
     """
     # Ask the K8s API for its version and check for errors.
-    url = f"{config.url}/version"
-    resp, err = get(client, url)
+    url = f"{k8sconfig.url}/version"
+    resp, err = get(k8sconfig.client, url)
     if err or resp is None:
-        return (None, True)
+        return (K8sConfig(), True)
 
     # Construct the version number of the K8s API.
     major, minor = resp['major'], resp['minor']
@@ -667,6 +641,240 @@ def version(config: K8sConfig, client) -> Tuple[Optional[K8sConfig], bool]:
     # to determines which URLs to contact, which fields are valid.
     version = version.replace("+", "")
 
-    # Return an updated `Config` tuple.
-    config = config._replace(version=version)
-    return (config, False)
+    # Return an updated `K8sconfig` tuple.
+    k8sconfig = k8sconfig._replace(version=version)
+    return (k8sconfig, False)
+
+
+def cluster_config(
+        kubeconfig: Filepath,
+        context: Optional[str]) -> Tuple[K8sConfig, bool]:
+    """Return web session to K8s API.
+
+    This will read the Kubernetes credentials, contact Kubernetes to
+    interrogate its version and then return the configuration and web-session.
+
+    Inputs:
+        kubeconfig: str
+            Path to kubeconfig file.
+        context: str
+            Kubernetes context to use (can be `None` to use default).
+
+    Returns:
+        K8sConfig
+
+    """
+    # Read Kubeconfig file and use it to create a `requests` client session.
+    # That session will have the proper security certificates and headers so
+    # that subsequent calls to K8s need not deal with it anymore.
+    kubeconfig = kubeconfig.expanduser()
+    try:
+        # Parse Kubeconfig file.
+        k8sconfig, err = load_auto_config(kubeconfig, context, disable_warnings=True)
+        assert not err
+
+        # Configure web session.
+        k8sconfig = k8sconfig._replace(client=session(k8sconfig))
+        assert k8sconfig.client
+
+        # Contact the K8s API to update version field in `k8sconfig`.
+        k8sconfig, err = version(k8sconfig)
+        assert not err and k8sconfig
+
+        # Populate the `k8sconfig.apis` field.
+        err = compile_api_endpoints(k8sconfig)
+        assert not err
+    except AssertionError:
+        return (K8sConfig(), True)
+
+    # Log the K8s API address and version.
+    logit.info(f"Kubernetes server at {k8sconfig.url}")
+    logit.info(f"Kubernetes version is {k8sconfig.version}")
+    return (k8sconfig, False)
+
+
+def parse_api_group(api_version, url, resp) -> Tuple[List[K8sResource], Dict[str, str]]:
+    """Compile the K8s API `resp` into a `K8sResource` tuples.
+
+    The `resp` is the verbatim response from the K8s API group regarding the
+    resources it provides. Here we compile those into `K8sResource` tuples iff
+    they meet the criteria to be manageable by Square. These criteria are, most
+    notably, the ability to create, get, patch and delete the resource.
+
+    Also return a LUT to convert short names like "svc" into the proper resource
+    kind "Service".
+
+    """
+    resources = resp["resources"]
+
+    def valid(_res):
+        """Return `True` if `res` describes a Square compatible resource."""
+        # Convenience.
+        name = _res["name"]
+        verbs = list(sorted(_res["verbs"]))
+
+        # Ignore resource like "services/status". We only care for "services".
+        if "/" in name:
+            logit.debug(f"Ignore resource <{name}>: has a slash ('/') in its name")
+            return False
+
+        # Square can only man age the resource if it can be read, modified and
+        # deleted. Here we check if `res` has the respective verbs.
+        minimal_verbs = {"create", "delete", "get", "list", "patch", "update"}
+        if not minimal_verbs.issubset(set(verbs)):
+            logit.debug(f"Ignore resource <{name}>: insufficient verbs: {verbs}")
+            return False
+
+        return True
+
+    # Compile the K8s resource definition into a `K8sResource` structure if it
+    # is compatible with Square (see `valid` helper above).
+    group_urls: List[K8sResource]
+    group_urls = [
+        K8sResource(api_version, _["kind"], _["name"], _["namespaced"], url)
+        for _ in resources if valid(_)
+    ]
+
+    # Compile LUT to translate short names into their proper resource
+    # kind: Example short2kind = {"service":, "Service", "svc": "Service"}
+    short2kind: Dict[str, str] = {}
+    for res in resources:
+        kind = res["kind"]
+
+        short2kind[kind.lower()] = kind
+        short2kind[res["name"]] = kind
+        for short_name in res.get("shortNames", []):
+            short2kind[short_name] = kind
+
+    return (group_urls, short2kind)
+
+
+def compile_api_endpoints(k8sconfig: K8sConfig) -> bool:
+    """Populate `k8sconfig.apis` with all the K8s endpoints`.
+
+    NOTE: This will purge the existing content in `k8sconfig.apis`.
+
+    Returns a dictionary like the following:
+    {
+      ('ConfigMap', 'v1'): K8sResource(
+        apiVersion=v1, kind='ConfigMap', name='configmaps', namespaced=True,
+        url='https://localhost:8443/api/v1/configmaps'),
+      ('CronJob', 'batch/v1beta1): K8sResource(
+        apiVersion='batch/v1beta1', kind='CronJob', name='cronjobs', namespaced=True,
+        url='https://localhost:8443/apis/batch/v1beta1/cronjobs'),
+      ('DaemonSet', 'apps/v1'): K8sResource(
+        apiVersion='apps/v1', kind='DaemonSet', name='daemonsets', namespaced=True,
+        url='https://localhost:8443/apis/apps/v1/daemonsets',
+      ('DaemonSet', apps/v1beta1): K8sResource(
+        apiVersion='apps/v1beta1', kind='DaemonSet', name='daemonsets', namespaced=True,
+        url='https://localhost:8443/apis/extensions/v1beta1/daemonsets'),
+    }
+
+    Inputs:
+        k8sconfig: K8sConfig
+
+    """
+    # Compile the list of all K8s API groups that this K8s instance knows about.
+    resp, err = get(k8sconfig.client, f"{k8sconfig.url}/apis")
+    if err:
+        logit.error(f"Could not interrogate the {k8sconfig.url}/apis")
+        return True
+
+    # Compile the list of all API groups and their endpoints. Example
+    # apigroups = {
+    #     'extensions': {('extensions/v1beta1', 'apis/extensions/v1beta1')},
+    #     'apps': {('apps/v1', 'apis/apps/v1'),
+    #              ('apps/v1beta1', 'apis/apps/v1beta1'),
+    #              ('apps/v1beta2', 'apis/apps/v1beta2')},
+    #     'batch': {('batch/v1', 'apis/batch/v1'),
+    #               ('batch/v1beta1', 'apis/batch/v1beta1')},
+    #     ...
+    # }
+    apigroups: Dict[str, Set[Tuple[str, str]]] = {}
+    preferred_group: Dict[str, str] = {}
+    for group in resp["groups"]:
+        name = group["name"]
+
+        # Store the preferred version, eg ("", "apis/v1").
+        apigroups[name] = set()
+
+        # Compile all alternative versions into the same set.
+        for version in group["versions"]:
+            ver = version["groupVersion"]
+            apigroups[name].add((ver, f"apis/{ver}"))
+            preferred_group[ver] = group["preferredVersion"]["groupVersion"]
+        del group
+
+    # The "v1" group comprises the traditional core components like Service and
+    # Pod. This group is a special case and exposed under "api/v1" instead
+    # of the usual `apis/...` path.
+    apigroups["v1"] = {("v1", "api/v1")}
+    preferred_group["v1"] = "v1"
+
+    # Contact K8s to find out which resources each API group offers.
+    # This will produce the following group_urls below (K = `K8sResource`): {
+    #  ('apps', 'apps/v1', 'apis/apps/v1'): [
+    #   K(*, kind='DaemonSet', name='daemonsets', namespaced=True, url='apis/apps/v1'),
+    #   K(*, kind='Deployment', name='deployments', namespaced=True, url='apis/apps/v1'),
+    #   K(*, kind='ReplicaSet', name='replicasets', namespaced=True, url='apis/apps/v1'),
+    #   K(*, kind='StatefulSet', name='statefulsets', namespaced=True, url='apis/apps/v1')
+    #  ],
+    #  ('apps', 'apps/v1beta1', 'apis/apps/v1beta1')': [
+    #    K(..., kind='Deployment', name='deployments', namespaced=True, url=...),
+    #    K(..., kind='StatefulSet', name='statefulsets', namespaced=True, url=...)
+    #  ],
+    # }
+    group_urls: Dict[Tuple[str, str, str], List[K8sResource]] = {}
+    for group_name, ver_url in apigroups.items():
+        for api_version, url in ver_url:
+            resp, err = get(k8sconfig.client, f"{k8sconfig.url}/{url}")
+            if err:
+                logit.error(f"Could not interrogate the {k8sconfig.url}/{url}")
+                return True
+
+            data, short2kind = parse_api_group(api_version, url, resp)
+            group_urls[(group_name, api_version, url)] = data
+            k8sconfig.short2kind.update(short2kind)
+
+    # Produce the entries for `K8sConfig.apis` as described in the doc string.
+    k8sconfig.apis.clear()
+    default: Dict[str, Set[K8sResource]] = defaultdict(set)
+    for (group_name, api_version, url), resources in group_urls.items():
+        for res in resources:
+            key = (res.kind, res.apiVersion)  # fixme: define namedtuple
+            k8sconfig.apis[key] = res._replace(url=f"{k8sconfig.url}/{res.url}")
+
+            if preferred_group[api_version] == api_version:
+                default[res.kind].add(k8sconfig.apis[key])
+                k8sconfig.apis[(res.kind, "")] = k8sconfig.apis[key]
+
+    # Determine the default API endpoint Square should query for each resource.
+    for kind, resources in default.items():
+        # Happy case: the resource is only available from a single API group.
+        if len(resources) == 1:
+            k8sconfig.apis[(kind, "")] = resources.pop()
+            continue
+
+        # If we get here then it means a resource is available from different
+        # API groups. Here we use heuristics to pick one. The heuristic is
+        # simply to look for one that is neither alpha nor beta. In Kubernetes
+        # v1.15 this resolves almost all disputes.
+        all_apis = list(sorted([_.apiVersion for _ in resources]))
+        logit.info(f"Ambiguous <{kind}> endpoints: {all_apis}")
+
+        # Remove all alpha/beta resources.
+        prod_apis = [_ for _ in all_apis if not ("alpha" in _ or "beta" in _)]
+
+        # Re-add the alpha/beta resources to the candidate set if we have no
+        # production ones to choose from.
+        apis = prod_apis if len(prod_apis) > 0 else list(all_apis)
+
+        # Pick the one with probably the highest version number.
+        apis.sort()
+        version = apis.pop()
+        k8sconfig.apis[(kind, "")] = [_ for _ in resources if _.apiVersion == version][0]
+
+    # Compile the set of all resource kinds that this Kubernetes cluster supports.
+    for kind, _ in k8sconfig.apis:
+        k8sconfig.kinds.add(kind)
+    return False
