@@ -12,8 +12,7 @@ import square
 import square.square
 from square import __version__
 from square.dtypes import (
-    RESOURCE_ALIASES, SUPPORTED_KINDS, Configuration, Filepath, GroupBy,
-    Selectors,
+    DEFAULT_PRIORITIES, Config, Filepath, GroupBy, Selectors,
 )
 
 # Convenience: global logger instance to avoid repetitive code.
@@ -32,26 +31,6 @@ def parse_commandline_args():
       {name} apply
     ''')
 
-    def _validate_kind(kind: str) -> str:
-        """Convert resource `kind` from aliases to canonical name.
-        For instance, `svc` -> `Service`.
-        """
-        kind = kind.lower()
-
-        # The "all" resource is special - do not expand.
-        if kind == "all":
-            return "all"
-
-        out = [
-            canonical for canonical, aliases in RESOURCE_ALIASES.items()
-            if kind in aliases
-        ]
-
-        # Must have found at most one or there is a serious bug.
-        if len(out) != 1:
-            raise argparse.ArgumentTypeError(kind)
-        return out[0]
-
     def _validate_label(label: str) -> Tuple[str, ...]:
         """Convert resource `kind` from aliases to canonical name.
         For instance, `svc` -> `Service`.
@@ -82,6 +61,11 @@ def parse_commandline_args():
         metavar="labels", dest="labels", default=tuple(),
         help="Only select resources with these labels (eg 'app=foo')",
     )
+    parent.add_argument(
+        "-p", "--priorities", nargs="*",
+        metavar="priorities", dest="priorities", default=tuple(DEFAULT_PRIORITIES),
+        help="Sort resource in this order when saving and applying",
+    )
     parent.add_argument(
         "--kubeconfig", type=str, metavar="path",
         default=os.environ.get("KUBECONFIG", None),
@@ -109,7 +93,7 @@ def parse_commandline_args():
     # duplicated code.
     kinds_kwargs = {
         "dest": "kinds",
-        "type": _validate_kind,
+        "type": str,
         "nargs": '+',
         "metavar": "resource",
     }
@@ -167,33 +151,43 @@ def user_confirmed(answer: Optional[str] = "yes") -> bool:
         return False
 
 
-def compile_config(cmdline_param) -> Tuple[Optional[Configuration], bool]:
-    """Return `Configuration` from `cmdline_param`.
+def compile_config(cmdline_param) -> Tuple[Config, bool]:
+    """Return `Config` from `cmdline_param`.
 
     Inputs:
         cmdline_param: SimpleNamespace
 
     Returns:
-        Configuration, err
+        Config, err
 
     """
+    err_resp = Config(
+        folder=Filepath(""),
+        kubeconfig=Filepath(""),
+        kube_ctx=None,
+        selectors=Selectors(set(), None, None),
+        groupby=GroupBy("", tuple()),
+        priorities=tuple(),
+    ), True
+
     # Convenience.
     p = cmdline_param
 
     # Abort without credentials.
-    if not p.kubeconfig:
-        logit.error("ERROR: must either specify --kubeconfig or set KUBECONFIG")
-        return None, True
+    kubeconfig = Filepath(p.kubeconfig)
+    if not kubeconfig.exists():
+        logit.error(f"Cannot find Kubernetes config file <{kubeconfig}>")
+        return err_resp
 
     # Remove duplicates but retain the original order of "p.kinds". This is
     # a "trick" that will only work in Python 3.7+ which guarantees a stable
     # insertion order for dictionaries (but not sets).
-    p.kinds = list(dict.fromkeys(p.kinds))
+    p.kinds = set(dict.fromkeys(p.kinds))
 
     # Expand the "all" resource (if present) and ignore all other resources, if
     # any were specified.
     if "all" in p.kinds:
-        p.kinds = list(SUPPORTED_KINDS)
+        p.kinds.clear()
 
     # Folder must be `Path` object.
     folder = pathlib.Path(p.folder)
@@ -211,12 +205,12 @@ def compile_config(cmdline_param) -> Tuple[Optional[Configuration], bool]:
     clean_order = [_ if not _.startswith("label") else "label" for _ in order]
     if not set(clean_order).issubset({"ns", "kind", "label"}):
         logit.error(f"Invalid resource names in <{order}>")
-        return None, True
+        return err_resp
 
     labels = [_ for _ in order if _.startswith("label")]
     if len(labels) > 1:
         logit.error("Can only specify one label in file hierarchy")
-        return None, True
+        return err_resp
 
     # Unpack the label name if the user specified it as part of `--groupby`.
     # Example, if user specified `--groupby label=app` then we need to extract
@@ -228,62 +222,47 @@ def compile_config(cmdline_param) -> Tuple[Optional[Configuration], bool]:
             assert len(label_name) > 0
         except (ValueError, AssertionError):
             logit.error(f"Invalid label specification <{labels[0]}>")
-            return None, True
-    groupby = GroupBy(order=clean_order, label=label_name)
+            return err_resp
+    groupby = GroupBy(order=tuple(clean_order), label=label_name)
     del order, clean_order, label_name
 
     # -------------------------------------------------------------------------
     # Assemble the full configuration and return it.
     # -------------------------------------------------------------------------
-    cfg = Configuration(
-        command=p.parser,
-        verbosity=p.verbosity,
+    cfg = Config(
         folder=folder,
-        kubeconfig=p.kubeconfig,
+        kubeconfig=kubeconfig,
         kube_ctx=p.ctx,
         selectors=selectors,
         groupby=groupby,
+        priorities=p.priorities,
     )
     return cfg, False
 
 
-def apply_plan(
-        kubeconfig: Filepath,
-        kube_ctx: Optional[str],
-        folder: Filepath,
-        selectors: Selectors,
-        confirm_string: Optional[str],
-) -> Tuple[None, bool]:
+def apply_plan(cfg: Config, confirm_string: Optional[str]) -> bool:
     """Update K8s to match the specifications in `local_manifests`.
 
     Create a deployment plan that will transition the K8s state
     `server_manifests` to the desired `local_manifests`.
 
     Inputs:
-        kubeconfig: Filepath,
-        kube_ctx: Optional[str],
-        folder: Filepath
-            Path to local manifests eg "./foo"
-        selectors: Selectors
-            Only operate on resources that match the selectors.
+        cfg: Square configuration.
         confirm_string:
             Only apply the plan if user answers with this string in the
             confirmation dialog (set to `None` to disable confirmation).
 
-    Returns:
-        None
-
     """
     try:
         # Obtain the plan.
-        plan, err = square.square.make_plan(kubeconfig, kube_ctx, folder, selectors)
+        plan, err = square.square.make_plan(cfg)
         assert not err and plan
 
         # Exit prematurely if there are no changes to apply.
         num_patch_ops = sum([len(_.patch.ops) for _ in plan.patch])
         if len(plan.create) == len(plan.delete) == num_patch_ops == 0:
             print("Nothing to change")
-            return (None, False)
+            return False
         del num_patch_ops
 
         # Print the plan and ask for user confirmation. Abort if the user does
@@ -291,17 +270,41 @@ def apply_plan(
         square.square.show_plan(plan)
         if not user_confirmed(confirm_string):
             print("User abort - no changes were made.")
-            return (None, True)
+            return True
         print()
 
         # Apply the plan.
-        _, err = square.square.apply_plan(kubeconfig, kube_ctx, plan)
-        assert not err
+        assert not square.square.apply_plan(cfg, plan)
     except AssertionError:
-        return (None, True)
+        return True
 
     # All good.
-    return (None, False)
+    return False
+
+
+def sanitise_resource_kinds(cfg: Config) -> Tuple[Config, bool]:
+    """Populate the `Selector.kinds` with those the user specified on the commandline."""
+    # Create a K8sConfig instance because it will contain all the info we need.
+    k8sconfig, err = square.k8s.cluster_config(cfg.kubeconfig, cfg.kube_ctx)
+    if err:
+        return (cfg, True)
+
+    # If the user did not specify any selectors then we assume he wants all resources.
+    if len(cfg.selectors.kinds) == 0:
+        cfg.selectors.kinds.update(k8sconfig.kinds)
+
+    # Translate colloquial names to their canonical counterpart if possible.
+    # Example: "svc" -> "Service"
+    # Do nothing if we do not recognise the name. This can happen because the
+    # user made a typo, or the resource is a custom resource that is not yet
+    # known. Since we cannot distinguish them, we allow typos and then ignore
+    # the resource during the get/plan/apply cycle.
+    kinds = {k8sconfig.short2kind.get(_.lower(), _) for _ in cfg.selectors.kinds}
+
+    # Replace the (possibly) colloquial names with their correct K8s kinds.
+    cfg.selectors.kinds.clear()
+    cfg.selectors.kinds.update(kinds)
+    return (cfg, False)
 
 
 def main() -> int:
@@ -314,21 +317,21 @@ def main() -> int:
     square.square.setup_logging(param.verbosity)
 
     # Create Square configuration from command line arguments.
-    cfg, err = compile_config(param)
-    if cfg is None or err:
+    cfg, err1 = compile_config(param)
+    cfg, err2 = sanitise_resource_kinds(cfg)
+    if err1 or err2:
         return 1
 
-    # Do what user asked us to do.
-    common_args = Filepath(cfg.kubeconfig), cfg.kube_ctx, cfg.folder, cfg.selectors
-    if cfg.command == "get":
-        _, err = square.square.get_resources(*common_args, cfg.groupby)
-    elif cfg.command == "plan":
-        plan, err = square.square.make_plan(*common_args)
+    # Do what the user asked us to do.
+    if param.parser == "get":
+        err = square.square.get_resources(cfg)
+    elif param.parser == "plan":
+        plan, err = square.square.make_plan(cfg)
         square.square.show_plan(plan)
-    elif cfg.command == "apply":
-        _, err = apply_plan(*common_args, "yes")
+    elif param.parser == "apply":
+        err = apply_plan(cfg, "yes")
     else:
-        logit.error(f"Unknown command <{cfg.command}>")
+        logit.error(f"Unknown command <{param.parser}>")
         return 1
 
     # Return error code.
