@@ -3,7 +3,9 @@ import copy
 import difflib
 import logging
 import pathlib
-from typing import DefaultDict, Dict, Iterable, List, Optional, Tuple
+from typing import (
+    Collection, DefaultDict, Dict, Iterable, List, Optional, Tuple,
+)
 
 import square.dotdict
 import square.k8s
@@ -11,9 +13,8 @@ import square.schemas
 import yaml
 import yaml.scanner
 from square.dtypes import (
-    NON_NAMESPACED_KINDS, SUPPORTED_KINDS, Filepath, GroupBy, K8sConfig,
-    LocalManifestLists, LocalManifests, MetaManifest, Selectors,
-    ServerManifests,
+    Config, Filepath, GroupBy, K8sConfig, LocalManifestLists, LocalManifests,
+    MetaManifest, Selectors, ServerManifests,
 )
 
 # Convenience: global logger instance to avoid repetitive code.
@@ -30,7 +31,7 @@ def make_meta(manifest: dict) -> MetaManifest:
     """
     # Unpack the namespace. For Namespace resources, this will be the "name".
     if manifest["kind"] == "Namespace":
-        ns = manifest['metadata']['name']
+        ns = None
     else:
         # For non-Namespace manifests, the namespace may genuinely be None if
         # the resource applies globally, eg ClusterRole.
@@ -96,11 +97,11 @@ def select(manifest: dict, selectors: Selectors) -> bool:
         logit.debug(f"Kind {kind} does not match selector {selectors.kinds}")
         return False
 
-    # Unless the namespace selector is None, the resource must match it.
-    if selectors.namespaces is not None:
-        if ns not in selectors.namespaces:
-            logit.debug(f"Namespace {ns} does not match selector {selectors.namespaces}")
-            return False
+    # Include the resource if it is a) namespaced b) we have namespace
+    # selectors and c) the resource matches one of them.
+    if ns and selectors.namespaces is not None and ns not in selectors.namespaces:
+        logit.debug(f"Namespace {ns} does not match selector {selectors.namespaces}")
+        return False
 
     # Convert the labels dictionary into a set of (key, value) tuples. We can
     # then use set logic to determine if the resource specifies the desired
@@ -118,7 +119,7 @@ def select(manifest: dict, selectors: Selectors) -> bool:
 
 
 def unpack_list(manifest_list: dict,
-                selectors: Selectors) -> Tuple[Optional[ServerManifests], bool]:
+                selectors: Selectors) -> Tuple[ServerManifests, bool]:
     """Unpack a K8s List item, eg `DeploymentList` or `NamespaceList`.
 
     Return a dictionary where each key uniquely identifies the resource via a
@@ -140,14 +141,14 @@ def unpack_list(manifest_list: dict,
     if len(missing) > 0:
         kind = manifest_list.get("kind", "UNKNOWN")
         logit.error(f"{kind} manifest is missing these keys: {missing}")
-        return (None, True)
+        return ({}, True)
     del must_have, missing
 
     # Sanity check: resource kind must end in "List", eg "DeploymentList".
     kind = manifest_list["kind"]
     if not kind.endswith('List'):
         logit.error(f"Kind {kind} is not a list")
-        return (None, True)
+        return ({}, True)
 
     # Strip of the "List".
     kind = kind[:-4]
@@ -172,7 +173,7 @@ def unpack_list(manifest_list: dict,
 
 def parse(
         file_yaml: Dict[Filepath, str],
-        selectors: Selectors) -> Tuple[Optional[LocalManifestLists], bool]:
+        selectors: Selectors) -> Tuple[LocalManifestLists, bool]:
     """Parse all YAML strings from `file_yaml` into `LocalManifestLists`.
 
     Exclude all manifests that do not satisfy the `selectors`.
@@ -202,7 +203,7 @@ def parse(
                 f"Cannot YAML parse <{fname}>"
                 f" - {err.problem} - Line {err.problem_mark.line}"
             )
-            return (None, True)
+            return ({}, True)
 
         # Remove all empty manifests. This typically happens when the YAML
         # file ends with a "---" string.
@@ -219,7 +220,7 @@ def parse(
             out[fname] = [(make_meta(_), _) for _ in manifests]
         except KeyError:
             logit.error(f"{file_yaml} does not look like a K8s manifest file.")
-            return None, True
+            return {}, True
 
     # Drop all files without manifests.
     out = {k: v for k, v in out.items() if len(v) > 0}
@@ -230,7 +231,7 @@ def parse(
     return (out, False)
 
 
-def unpack(manifests: LocalManifestLists) -> Tuple[Optional[ServerManifests], bool]:
+def unpack(manifests: LocalManifestLists) -> Tuple[ServerManifests, bool]:
     """Convert `manifests` to `ServerManifests` for internal processing.
 
     Returns `False` unless all resources in `manifests` are unique. For
@@ -267,82 +268,17 @@ def unpack(manifests: LocalManifestLists) -> Tuple[Optional[ServerManifests], bo
                 f"Defined in {str.join(', ', tmp)}"
             )
     if not unique:
-        return (None, True)
+        return ({}, True)
 
     # Compile the input manifests into a new dict with the meta manifest as key.
     out = {k: v for fname in manifests for k, v in manifests[fname]}
     return (out, False)
 
 
-def unparse(
-        file_manifests: LocalManifestLists
-) -> Tuple[Optional[Dict[Filepath, str]], bool]:
-    """Convert the Python dict to a Yaml string for each file and return it.
-
-    The output dict can be passed directly to `save_files` to write the files.
-
-    Inputs:
-        file_manifests: Dict[Filepath:Tuple[MetaManifest, manifest]]
-            Typically the output from eg `manio.sync`.
-
-    Returns:
-        Dict[Filepath:YamlStr]: Yaml representation of all manifests.
-
-    """
-    out = {}
-    for fname, manifests in file_manifests.items():
-        # Verify that this file contains only supported resource kinds.
-        kinds = {meta.kind for meta, _ in manifests}
-        delta = kinds - set(SUPPORTED_KINDS)
-        if len(delta) > 0:
-            logit.error(f"Found unsupported KIND when writing <{fname}>: {delta}")
-            return (None, True)
-
-        # Group the manifests by their "kind", sort each group and compile a
-        # new list of grouped and sorted manifests.
-        man_sorted: List[dict] = []
-        for kind in SUPPORTED_KINDS:
-            man_sorted += sorted([_ for _ in manifests if _[0].kind == kind])
-        assert len(man_sorted) == len(manifests)
-
-        # Drop the MetaManifest, ie
-        # Dict[Filepath:Tuple[MetaManifest, manifest]] -> Dict[Filepath:manifest]
-        man_clean = [manifest for _, manifest in man_sorted]
-
-        # Assign the grouped and sorted list of manifests to the output dict.
-        out[fname] = man_clean
-        del fname, manifests, man_sorted, man_clean
-
-    # Ignore all files whose manifest list is empty.
-    out_nonempty = {k: v for k, v in out.items() if len(v) > 0}
-    del out
-
-    # Ensure that all dicts are pure Python dicts or there will be problems
-    # with the YAML generation below.
-    out_clean = {k: square.dotdict.undo(v) for k, v in out_nonempty.items()}
-    del out_nonempty
-
-    # Convert all manifest dicts into YAML strings.
-    out_final: Dict[Filepath, str] = {}
-    try:
-        for fname, v in out_clean.items():
-            out_final[fname] = yaml.safe_dump_all(v, default_flow_style=False)
-    except yaml.YAMLError as err:
-        logit.error(
-            f"YAML error. Cannot create <{fname}>: {err.args[0]} <{str(err.args[1])}>"
-        )
-        return (None, True)
-
-    # Return the Dict[Filepath:YamlStr]
-    return (out_final, False)
-
-
-def sync(
-        local_manifests: LocalManifestLists,
-        server_manifests: ServerManifests,
-        selectors: Selectors,
-        groupby: GroupBy,
-) -> Tuple[Optional[LocalManifestLists], bool]:
+def sync(local_manifests: LocalManifestLists,
+         server_manifests: ServerManifests,
+         selectors: Selectors,
+         groupby: GroupBy) -> Tuple[LocalManifestLists, bool]:
     """Update the local manifests with the server values and return the result.
 
     Inputs:
@@ -357,12 +293,6 @@ def sync(
         Dict[Filepath, Tuple[MetaManifest, dict]]
 
     """
-    # Sanity check: all `kinds` must be supported or we abort.
-    if not set(selectors.kinds).issubset(SUPPORTED_KINDS):
-        unsupported = set(selectors.kinds) - set(SUPPORTED_KINDS)
-        logit.error(f"Cannot sync unsupported kinds: {unsupported}")
-        return (None, True)
-
     # Avoid side effects.
     server_manifests = copy.deepcopy(server_manifests)
 
@@ -410,7 +340,7 @@ def sync(
         except KeyError:
             fname, err = filename_for_manifest(meta, manifest, groupby)
             if err:
-                return (None, True)
+                return ({}, True)
             out_add_mod[fname].append((meta, manifest))
         else:
             # Update the correct YAML document in the correct file.
@@ -459,10 +389,11 @@ def filename_for_manifest(
     # "--groupby" command line option accepts. We will use this LUT below to
     # assemble the full manifest path.
     lut = {
-        # Get the namespace. Use "_global_" if the resource's namespace is
-        # `None`, which it will be for global resources like ClusterRole and
-        # ClusterRoleBinding.
-        "ns": meta.namespace or "_global_",
+        # Get the namespace. Use "_global_" for non-namespaced resources.
+        # The only exception are `Namespaces` themselves because it is neater
+        # to save their manifest in the relevant namespace folder, together
+        # with all the other resources that are in that namespace.
+        "ns": (meta.name if meta.kind == "Namespace" else None) or meta.namespace or "_global_",  # noqa
         "kind": meta.kind.lower(),
         # Try to find the user specified label. If the current resource lacks
         # that label then put it into the catchall file.
@@ -481,17 +412,18 @@ def filename_for_manifest(
     return Filepath(path), False
 
 
-def diff(
-        config: K8sConfig,
-        local: LocalManifests,
-        server: ServerManifests) -> Tuple[Optional[str], bool]:
+def diff(config: Config,
+         k8sconfig: K8sConfig,
+         local: LocalManifests,
+         server: ServerManifests) -> Tuple[str, bool]:
     """Return the human readable diff between the `local` and `server`.
 
     The diff shows the necessary changes to transition the `server` manifest
     into the state of the `local` manifest.
 
     Inputs:
-        config: K8sConfig
+        config: Square configuration.
+        k8sconfig: K8sConfig
         local: dict
             Local manifest.
         server: dict
@@ -504,10 +436,10 @@ def diff(
     """
     # Clean up the input manifests because we do not want to diff, for instance,
     # the `status` fields.
-    (srv, _), err1 = strip(config, server)
-    (loc, _), err2 = strip(config, local)
+    (srv, _), err1 = strip(k8sconfig, server, square.schemas.EXCLUSION_SCHEMA)
+    (loc, _), err2 = strip(k8sconfig, local, square.schemas.EXCLUSION_SCHEMA)
     if err1 or err2:
-        return (None, True)
+        return ("", True)
 
     # Undo the DotDicts. This is a precaution because the YAML parser can
     # otherwise not dump the manifests.
@@ -522,55 +454,38 @@ def diff(
 
 
 def strip(
-        config: K8sConfig,
-        manifest: dict
+        k8sconfig: K8sConfig,
+        manifest: dict,
+        exclusion_schema: Dict[str, dict],
         ) -> Tuple[Tuple[DotDict, dict], bool]:
     """Strip `manifest` according to the exclusion filters in `square.schemas`.
 
     Inputs:
-        config: K8sConfig
+        k8sconfig: K8sConfig
         manifest: dict
+        exclusion_schema: Dict[str, dict]
+            fixme: docu
 
     Returns:
         dict: the removed keys.
 
     """
-    assert config.version is not None
+    assert k8sconfig.version is not None
 
     # Convenience: default return value if an error occurs.
     ret_err: Tuple[Tuple[DotDict, dict], bool] = ((square.dotdict.make({}), {}), True)
 
-    # Avoid side effects.
-    manifest = copy.deepcopy(manifest)
-
-    # Every manifest must specify its "apiVersion" and "kind".
+    # Parse the manifest.
     try:
-        kind = manifest["kind"]
-    except KeyError as err:
-        logit.error(f"Manifest is missing the <{err.args[0]}> key.")
+        meta = make_meta(manifest)
+        resource, err = square.k8s.resource(k8sconfig, meta)
+        assert not err
+    except KeyError as e:
+        logit.error(f"Manifest is missing the <{e.args[0]}> key.")
         return ret_err
-
-    # Unpack the name and namespace to produce a convenient log message.
-    # NOTE: we assume here that manifests may not have either.
-    name = manifest.get("metadata", {}).get("name", None)
-    ns = manifest.get("metadata", {}).get("namespace", None)
-
-    # Abort if manifest lacks `metadata.name`.
-    if name is None:
-        logit.error("<metadata.name> must not be empty")
+    except AssertionError:
         return ret_err
 
-    # All but cluster level resources must have a `metadata.namespace` field.
-    if kind in NON_NAMESPACED_KINDS:
-        if ns is not None:
-            logit.error(f"<{kind}> must not have a <metadata.namespace> field.")
-            return ret_err
-    else:
-        if ns is None:
-            logit.error(f"<{kind}> must have <metadata.namespace> field.")
-            return ret_err
-    del name, ns
-
     def _update(exclude: dict, manifest: dict):
         """Recursively traverse the `manifest` and prune it according to `exclude`.
 
@@ -603,12 +518,14 @@ def strip(
         # Remove all empty sub-dictionaries from `removed`.
         return {k: v for k, v in removed.items() if v != {}}
 
-    # Verify the exclusion schema for the current resource and K8s version exist.
-    try:
-        exclude = square.schemas.EXCLUSION_SCHEMA[manifest["kind"]]
-    except KeyError:
-        logit.error(f"Unknown resource kind: <{kind}>")
-        return ret_err
+    # Add a default exclusion scheme for the current resource if none exists yet.
+    if manifest["kind"] not in exclusion_schema:
+        logit.warning(f"No exclusion schema for: <{meta.kind}> - using default")
+        exclusion_schema[manifest["kind"]] = {}
+        square.schemas.populate_schemas(exclusion_schema)
+
+    # Get the exclusion scheme for the current resource.
+    exclude = exclusion_schema[manifest["kind"]]
 
     # Strip down the manifest to its essential parts and return it.
     manifest = copy.deepcopy(manifest)
@@ -738,7 +655,7 @@ def align_serviceaccount(
     return (local_manifests, False)
 
 
-def save_files(folder: Filepath, file_data: Dict[Filepath, str]) -> Tuple[None, bool]:
+def save_files(folder: Filepath, file_data: Dict[Filepath, str]) -> bool:
     """Save all `file_data` under `folder`.
 
     All paths in `file_data` are relative to `folder`.
@@ -762,7 +679,7 @@ def save_files(folder: Filepath, file_data: Dict[Filepath, str]) -> Tuple[None,
             fp.unlink()
     except (IOError, PermissionError) as err:
         logit.error(f"{err}")
-        return (None, True)
+        return True
 
     # Iterate over the dict and write each file. Abort on error.
     for fname, yaml_str in file_data.items():
@@ -781,15 +698,15 @@ def save_files(folder: Filepath, file_data: Dict[Filepath, str]) -> Tuple[None,
             fname_abs.write_text(yaml_str)
         except (IOError, PermissionError) as err:
             logit.error(f"{err}")
-            return (None, True)
+            return True
 
     # Tell caller that all files were successfully written.
-    return (None, False)
+    return False
 
 
 def load_files(
         folder: Filepath,
-        fnames: Iterable[Filepath]) -> Tuple[Optional[Dict[Filepath, str]], bool]:
+        fnames: Iterable[Filepath]) -> Tuple[Dict[Filepath, str], bool]:
     """Load all `fnames` in `folder` and return their content.
 
     The elements of `fname` can have sub-paths, eg `foo/bar/file.txt` is valid
@@ -824,14 +741,14 @@ def load_files(
             out[fname_rel] = fname_abs.read_text()
         except FileNotFoundError:
             logit.error(f"Could not find <{fname_abs}>")
-            return (None, True)
+            return ({}, True)
 
     # Return the read files.
     return (out, False)
 
 
 def load(folder: Filepath, selectors: Selectors) -> Tuple[
-        Optional[ServerManifests], Optional[LocalManifestLists], bool]:
+        ServerManifests, LocalManifestLists, bool]:
     """Recursively load all "*.yaml" files under `folder`.
 
     Ignores all files not ending in ".yaml". Also removes all manifests that do
@@ -871,44 +788,113 @@ def load(folder: Filepath, selectors: Selectors) -> Tuple[
         man_meta, err = unpack(man_files)
         assert not err and man_meta is not None
     except AssertionError:
-        return (None, None, True)
+        return ({}, {}, True)
 
     # Return the file based manifests and unpacked manifests.
     return (man_meta, man_files, False)
 
 
-def save(folder: Filepath, manifests: LocalManifestLists) -> Tuple[None, bool]:
-    """Convert all `manifests` to YAML and save them.
+def sort_manifests(
+        file_manifests: LocalManifestLists,
+        priority: Collection[str]
+) -> Tuple[Dict[Filepath, Collection[MetaManifest]], bool]:
+    """Sort the manifests in each `file_manifests` by their `priority`.
 
-    Returns no data in the case of an error.
+    The returned data contains only the manifests without the `MetaData`. The
+    idea is to pass that data directly to `save_files`.
 
-    NOTE: this is merely a wrapper around the various low-level functions to
-    create YAML string and save the files.
+    Inputs:
+        file_manifests: the manifests that should go into each file.
+        priority: Collection[str]
+            Sort the manifest in this order, or alphabetically at the end if
+            not in the list.
+
+    """
+    # Sort the manifests in each file.
+    out: Dict[Filepath, Collection[MetaManifest]] = {}
+    for fname, manifests in file_manifests.items():
+        # Group the manifests by their "kind" in order of `priority` and sort
+        # each group alphabetically.
+        man_sorted: List[dict] = []
+        for kind in priority:
+            # Partition the manifest list into the current `kind` and the rest.
+            tmp = [_ for _ in manifests if _[0].kind == kind]
+            manifests = [_ for _ in manifests if _[0].kind != kind]
+
+            # Append the manifests ordered by their MetaManifest.
+            man_sorted += sorted(tmp, key=lambda _: _[0])
+
+        # Group the remaining manifests by their "kind" and sort each group
+        # alphabetically.
+        remaining_kinds = {_[0].kind for _ in manifests}
+        for kind in sorted(remaining_kinds):
+            # Partition the manifest list into the current `kind` and the rest.
+            tmp = [_ for _ in manifests if _[0].kind == kind]
+            manifests = [_ for _ in manifests if _[0].kind != kind]
+
+            # Append the manifests ordered by their MetaManifest.
+            man_sorted += sorted(tmp, key=lambda _: _[0])
+
+        # sanity check: we must have used up all manifests.
+        assert len(manifests) == 0
+
+        # Drop the MetaManifest, ie
+        # Dict[Filepath:Tuple[MetaManifest, manifest]] -> Dict[Filepath:manifest]
+        man_clean = [manifest for _, manifest in man_sorted]
+
+        # Assign the grouped and sorted list of manifests to the output dict.
+        out[fname] = man_clean
+
+    return out, False
+
+
+def save(folder: Filepath, manifests: LocalManifestLists,
+         priority: Collection[str]) -> bool:
+    """Saves all `manifests` as YAMLs in `folder`.
 
     Input:
         folder: Filepath
             Source folder.
         file_manifests: Dict[Filepath, Tuple(MetaManifest, dict)]
             Names of files and their Python dicts to save as YAML.
+        priority: Collection[str]
+            Sort the manifest in this order, or alphabetically at the end if
+            not in the list.
 
     Returns:
         None
 
     """
-    # Convert the manifest to YAML strings. Abort on error.
-    fdata_raw, err = unparse(manifests)
-    if err or fdata_raw is None:
-        return (None, True)
+    # Sort the manifest in each file by priority.
+    out, err = sort_manifests(manifests, priority)
+    if err:
+        return True
+
+    # Ignore all files without manifests, ie empty files.
+    out_nonempty = {k: v for k, v in out.items() if len(v) > 0}
+    del out
+
+    # Ensure we have no `DotDicts` left. This will avoid problems with the YAML
+    # serialisation below.
+    out_clean = {k: square.dotdict.undo(v) for k, v in out_nonempty.items()}
+    del out_nonempty
+
+    # Convert all manifest dicts into YAML strings.
+    out_final: Dict[Filepath, str] = {}
+    try:
+        for fname, v in out_clean.items():
+            out_final[fname] = yaml.safe_dump_all(v, default_flow_style=False)
+    except yaml.YAMLError as e:
+        logit.error(
+            f"YAML error. Cannot create <{fname}>: {e.args[0]} <{str(e.args[1])}>"
+        )
+        return True
 
     # Save the files to disk.
-    return save_files(folder, fdata_raw)
+    return save_files(folder, out_final)
 
 
-def download(
-        config: K8sConfig,
-        client,
-        selectors: Selectors,
-) -> Tuple[Optional[ServerManifests], bool]:
+def download(config: Config, k8sconfig: K8sConfig) -> Tuple[ServerManifests, bool]:
     """Download and return the resources that match `selectors`.
 
     Set `selectors.namespace` to `None` to download the resources from all
@@ -917,9 +903,8 @@ def download(
     Either returns all the data or an error; never returns partial results.
 
     Inputs:
-        config: K8sConfig
-        client: `requests` session with correct K8s certificates.
-        selectors: Selectors
+        config: Square configuration.
+        k8sconfig: K8sConfig
 
     Returns:
         Dict[MetaManifest, dict]: the K8s manifests from K8s.
@@ -930,30 +915,35 @@ def download(
 
     # Ensure `namespaces` is always a list to avoid special casing below.
     all_namespaces: Iterable[Optional[str]]
-    if selectors.namespaces is None:
+    if config.selectors.namespaces is None:
         all_namespaces = [None]
     else:
-        all_namespaces = selectors.namespaces
+        all_namespaces = config.selectors.namespaces
 
     # Download each resource type. Abort at the first error and return nothing.
     for namespace in all_namespaces:
-        for kind in selectors.kinds:
-            try:
-                # Get the HTTP URL for the resource request.
-                url, err = square.k8s.urlpath(config, kind, namespace)
-                assert not err and url is not None
+        for kind in sorted(config.selectors.kinds):
+            # Get the K8s URL for the current resource kind. Ignore this
+            # resource if K8s does not know about it. The reason for that could
+            # by a typo or that it is a custom resource that does not (yet) exist.
+            resource, err = square.k8s.resource(k8sconfig, MetaManifest("", kind, namespace, ""))  # noqa
+            if err:
+                logit.warning(f"Skipping unknown resource <{kind}>")
+                continue
 
-                # Make HTTP request.
-                manifest_list, err = square.k8s.get(client, url)
+            try:
+                # Download the resource list from K8s.
+                manifest_list, err = square.k8s.get(k8sconfig.client, resource.url)
                 assert not err and manifest_list is not None
 
                 # Parse the K8s List (eg DeploymentList, NamespaceList, ...) into a
                 # Dict[MetaManifest, dict] dictionary.
-                manifests, err = unpack_list(manifest_list, selectors)
+                manifests, err = unpack_list(manifest_list, config.selectors)
                 assert not err and manifests is not None
 
                 # Drop all manifest fields except "apiVersion", "metadata" and "spec".
-                ret = {k: strip(config, man) for k, man in manifests.items()}
+                ret = {k: strip(k8sconfig, man, square.schemas.EXCLUSION_SCHEMA)
+                       for k, man in manifests.items()}
 
                 # Ensure `strip` worked for every manifest.
                 err = any((v[1] for v in ret.values()))
@@ -966,7 +956,7 @@ def download(
                 manifests = {k: v[0][0] for k, v in ret.items() if v[0][0] is not None}
             except AssertionError:
                 # Return nothing, even if we had downloaded other kinds already.
-                return (None, True)
+                return ({}, True)
             else:
                 # Copy the manifests into the output dictionary.
                 server_manifests.update(manifests)
