@@ -1,17 +1,18 @@
 import json
 import logging
-from typing import Any, Iterable, Optional, Set, Tuple
+from typing import Collection, Optional, Set, Tuple
 
 import colorama
 import jsonpatch
 import square.k8s as k8s
 import square.manio as manio
+import square.schemas
 import yaml
 from colorlog import ColoredFormatter
 from square.dtypes import (
-    NON_NAMESPACED_KINDS, SUPPORTED_KINDS, DeltaCreate, DeltaDelete,
-    DeltaPatch, DeploymentPlan, DeploymentPlanMeta, Filepath, GroupBy,
-    JsonPatch, K8sConfig, MetaManifest, Selectors, ServerManifests,
+    Config, DeltaCreate, DeltaDelete, DeltaPatch, DeploymentPlan,
+    DeploymentPlanMeta, JsonPatch, K8sConfig, MetaManifest, Selectors,
+    ServerManifests,
 )
 
 # Convenience: global logger instance to avoid repetitive code.
@@ -19,12 +20,15 @@ logit = logging.getLogger("square")
 
 
 def make_patch(
-        config: K8sConfig,
+        config: Config,
+        k8sconfig: K8sConfig,
         local: ServerManifests,
-        server: ServerManifests) -> Tuple[Optional[JsonPatch], bool]:
+        server: ServerManifests) -> Tuple[JsonPatch, bool]:
     """Return JSON patch to transition `server` to `local`.
 
     Inputs:
+        config: Square configuration.
+        k8sconfig: K8sConfig
         local: LocalManifests
             Usually the dictionary keys returned by `load_manifest`.
         server: ServerManifests
@@ -36,10 +40,10 @@ def make_patch(
     """
     # Reduce local and server manifests to salient fields (ie apiVersion, kind,
     # metadata and spec). Abort on error.
-    (loc, _), err1 = manio.strip(config, local)
-    (srv, _), err2 = manio.strip(config, server)
+    (loc, _), err1 = manio.strip(k8sconfig, local, square.schemas.EXCLUSION_SCHEMA)
+    (srv, _), err2 = manio.strip(k8sconfig, server, square.schemas.EXCLUSION_SCHEMA)
     if err1 or err2 or loc is None or srv is None:
-        return (None, True)
+        return (JsonPatch("", []), True)
 
     # Log the manifest info for which we will try to compute a patch.
     man_id = f"{loc.kind.upper()}: {loc.metadata.name}/{loc.metadata.name}"
@@ -47,19 +51,10 @@ def make_patch(
 
     # Sanity checks: abort if the manifests do not specify the same resource.
     try:
-        assert srv.apiVersion == loc.apiVersion
-        assert srv.kind == loc.kind
-        assert srv.metadata.name == loc.metadata.name
-
-        # Not all resources live in a namespace, for instance Namespaces,
-        # ClusterRoles, ClusterRoleBindings. Here we ensure that the namespace
-        # in the local and server manifest matches for those resources that
-        # have a namespace.
-        if srv.kind.lower() in {_.lower() for _ in NON_NAMESPACED_KINDS}:
-            namespace = None
-        else:
-            assert srv.metadata.namespace == loc.metadata.namespace
-            namespace = srv.metadata.namespace
+        res_srv, err_srv = k8s.resource(k8sconfig, manio.make_meta(srv))
+        res_loc, err_loc = k8s.resource(k8sconfig, manio.make_meta(loc))
+        assert err_srv is err_loc is False
+        assert res_srv == res_loc
     except AssertionError:
         # Log the invalid manifests and return with an error.
         keys = ("apiVersion", "kind", "metadata")
@@ -69,20 +64,14 @@ def make_patch(
             "Cannot compute JSON patch for incompatible manifests. "
             f"Local: <{loc_tmp}>  Server: <{srv_tmp}>"
         )
-        return (None, True)
-
-    # Determine the PATCH URL for the resource.
-    url, err = k8s.urlpath(config, srv.kind, namespace)
-    if err:
-        return (None, True)
-    full_url = f'{url}/{srv.metadata.name}'
+        return (JsonPatch("", []), True)
 
     # Compute JSON patch.
     patch = jsonpatch.make_patch(srv, loc)
     patch = json.loads(patch.to_string())
 
     # Return the patch.
-    return (JsonPatch(full_url, patch), False)
+    return (JsonPatch(res_srv.url, patch), False)
 
 
 def partition_manifests(
@@ -128,10 +117,60 @@ def partition_manifests(
     return (plan, False)
 
 
+def preferred_api(k8sconfig: K8sConfig,
+                  local: ServerManifests) -> Tuple[ServerManifests, bool]:
+    """Return a new version of `local` where all APIs point to the preferred group.
+
+    Example:
+      in = {
+        MetaManifest(apiVersion='extensions/v1beta1', kind='Ingress', ...): {
+          'apiVersion': 'networking.k8s.io/v1beta1',
+          ...
+        }
+      }
+      out = {
+        MetaManifest(apiVersion='networking.k8s.io/v1beta1', kind='Ingress', ...): {
+          'apiVersion': 'networking.k8s.io/v1beta1',
+          ...
+        }
+      }
+
+    """
+    # Check the API version in each MetaManifest and update it to the preferred
+    # version in both the MetaManifest and the associated K8s YAML manifest.
+    out = {}
+    for meta, manifest in local.items():
+        # The current `meta` can be outdated but must still be a valid resource
+        # in the current cluster.
+        if k8s.resource(k8sconfig, meta)[1]:
+            return {}, True
+
+        # Get the canonical resource description.
+        res, err = k8s.resource(k8sconfig, meta._replace(apiVersion=""))
+        if err:
+            logit.critical(f"BUG: resource <{meta}> should have been valid")
+            return {}, True
+
+        # Log a warning if the manifests uses an outdated API group.
+        if meta.apiVersion != res.apiVersion:
+            logit.warning(
+                f"Switching <{res.kind} {meta.namespace}/{meta.name}>"
+                f" from <{meta.apiVersion}> to <{res.apiVersion}>"
+                " - Please update your manifests."
+            )
+
+            # Update the API version for the resource to whatever K8s prefers.
+            meta = meta._replace(apiVersion=res.apiVersion)
+            manifest["apiVersion"] = meta.apiVersion
+        out[meta] = manifest
+    return out, False
+
+
 def compile_plan(
-        config: K8sConfig,
+        config: Config,
+        k8sconfig: K8sConfig,
         local: ServerManifests,
-        server: ServerManifests) -> Tuple[Optional[DeploymentPlan], bool]:
+        server: ServerManifests) -> Tuple[DeploymentPlan, bool]:
     """Return the `DeploymentPlan` to transition K8s to state of `local`.
 
     The deployment plan is a named tuple. It specifies which resources to
@@ -139,7 +178,8 @@ def compile_plan(
     specified in `local`.
 
     Inputs:
-        config: K8sConfig
+        config: Square configuration.
+        k8sconfig: K8sConfig
         local: ServerManifests
             Should be output from `load_manifest` or `load`.
         server: ServerManifests
@@ -149,10 +189,17 @@ def compile_plan(
         DeploymentPlan
 
     """
+    err_resp = (DeploymentPlan(tuple(), tuple(), tuple()), True)
+
+    # Replace the API group of the local resource with the one K8s prefers.
+    local, err = preferred_api(k8sconfig, local)
+    if err:
+        return err_resp
+
     # Partition the set of meta manifests into create/delete/patch groups.
     plan, err = partition_manifests(local, server)
     if err or not plan:
-        return (None, True)
+        return err_resp
 
     # Sanity check: the resources to patch *must* exist in both local and
     # server manifests. This is a bug if not.
@@ -162,10 +209,12 @@ def compile_plan(
     # Compile the Deltas to create the missing resources.
     create = []
     for delta in plan.create:
-        url, err = k8s.urlpath(config, delta.kind, namespace=delta.namespace)
-        if err or not url:
-            return (None, True)
-        create.append(DeltaCreate(delta, url, local[delta]))
+        # We only need the resource and namespace, not its name, because that
+        # is how the POST request to create a resource works in K8s.
+        resource, err = k8s.resource(k8sconfig, delta._replace(name=""))
+        if err or not resource:
+            return err_resp
+        create.append(DeltaCreate(delta, resource.url, local[delta]))
 
     # Compile the Deltas to delete the excess resources. Every DELETE request
     # will have to pass along a `DeleteOptions` manifest (see below).
@@ -178,15 +227,12 @@ def compile_plan(
     delete = []
     for meta in plan.delete:
         # Resource URL.
-        url, err = k8s.urlpath(config, meta.kind, namespace=meta.namespace)
-        if err or not url:
-            return (None, True)
-
-        # DELETE requests must specify the resource name in the path.
-        url = f"{url}/{meta.name}"
+        resource, err = k8s.resource(k8sconfig, meta)
+        if err or not resource:
+            return err_resp
 
         # Compile the Delta and add it to the list.
-        delete.append(DeltaDelete(meta, url, del_opts.copy()))
+        delete.append(DeltaDelete(meta, resource.url, del_opts.copy()))
 
     # Iterate over each manifest that needs patching and determine the
     # necessary JSON Patch to transition K8s into the state specified in the
@@ -194,15 +240,15 @@ def compile_plan(
     patches = []
     for meta in plan.patch:
         # Compute textual diff (only useful for the user to study the diff).
-        diff_str, err = manio.diff(config, local[meta], server[meta])
+        diff_str, err = manio.diff(config, k8sconfig, local[meta], server[meta])
         if err or diff_str is None:
-            return (None, True)
+            return err_resp
 
         # Compute the JSON patch that will change the K8s state to match the
         # one in the local files.
-        patch, err = make_patch(config, local[meta], server[meta])
+        patch, err = make_patch(config, k8sconfig, local[meta], server[meta])
         if err or patch is None:
-            return (None, True)
+            return err_resp
 
         # Append the patch to the list of patches, unless it is empty.
         if len(patch.ops):
@@ -212,7 +258,7 @@ def compile_plan(
     return (DeploymentPlan(create, patches, delete), False)
 
 
-def show_plan(plan: Optional[DeploymentPlan]) -> Tuple[None, bool]:
+def show_plan(plan: Optional[DeploymentPlan]) -> bool:
     """Print human readable version of `plan` to terminal.
 
     Inputs:
@@ -225,7 +271,7 @@ def show_plan(plan: Optional[DeploymentPlan]) -> Tuple[None, bool]:
     # Do nothing if the plan is `None`. This special case makes it easier to
     # deal with cases where `square.make_plan` returns an error.
     if not plan:
-        return (None, False)
+        return False
 
     # Terminal colours for convenience.
     cAdd = colorama.Fore.GREEN
@@ -238,6 +284,7 @@ def show_plan(plan: Optional[DeploymentPlan]) -> Tuple[None, bool]:
     # Use Green to list all the resources that we should create.
     for delta_c in plan.create:
         name = f"{delta_c.meta.kind.upper()} {delta_c.meta.namespace}/{delta_c.meta.name}"
+        name += f" ({delta_c.meta.apiVersion})"
 
         # Convert manifest to YAML string and print every line in Green.
         txt = yaml.dump(delta_c.manifest, default_flow_style=False)
@@ -270,12 +317,14 @@ def show_plan(plan: Optional[DeploymentPlan]) -> Tuple[None, bool]:
         formatted_diff = str.join('\n', colour_lines)
 
         name = f"{delta_p.meta.kind.upper()} {delta_p.meta.namespace}/{delta_p.meta.name}"
+        name += f" ({delta_p.meta.apiVersion})"
         print(cMod + f"Patch {name}" + cReset + "\n" + formatted_diff + "\n")
         n_mod += 1
 
     # Use Red to list all the resources that we should delete.
     for delta_d in plan.delete:
         name = f"{delta_d.meta.kind.upper()} {delta_d.meta.namespace}/{delta_d.meta.name}"
+        name += f" ({delta_d.meta.apiVersion})"
         print(cDel + f"Delete {name}" + cReset)
         n_del += 1
 
@@ -285,12 +334,12 @@ def show_plan(plan: Optional[DeploymentPlan]) -> Tuple[None, bool]:
           cMod + f"{n_mod:,} to change, " +  # noqa
           cDel + f"{n_del:,} to destroy." +  # noqa
           cReset + "\n")
-    return (None, False)
+    return False
 
 
 def find_namespace_orphans(
-        meta_manifests: Iterable[MetaManifest]
-) -> Tuple[Optional[Set[MetaManifest]], bool]:
+        meta_manifests: Collection[MetaManifest]
+) -> Tuple[Set[MetaManifest], bool]:
     """Return all orphaned resources in the `meta_manifest` set.
 
     A resource is orphaned iff it lives in a namespace that is not explicitly
@@ -301,7 +350,7 @@ def find_namespace_orphans(
     their namespace field.
 
     Inputs:
-        meta_manifests: Iterable[MetaManifest]
+        meta_manifests: Collection[MetaManifest]
 
     Returns:
         set[MetaManifest]: orphaned resources.
@@ -368,16 +417,11 @@ def setup_logging(log_level: int) -> None:
     logit.info(f"Set log level to {level}")
 
 
-def apply_plan(
-        kubeconfig: Filepath,
-        kube_ctx: Optional[str],
-        plan: DeploymentPlan,
-) -> Tuple[None, bool]:
+def apply_plan(cfg: Config, plan: DeploymentPlan) -> bool:
     """Update K8s resources according to the `plan`.
 
     Inputs:
-        kubeconfig: Filepath,
-        kube_ctx: Optional[str],
+        cfg: Square configuration.
         plan: DeploymentPlan
 
     Returns:
@@ -386,14 +430,14 @@ def apply_plan(
     """
     try:
         # Create properly configured Requests session to talk to K8s API.
-        (k8s_config, k8s_client), err = cluster_config(kubeconfig, kube_ctx)
-        assert not err and k8s_config and k8s_client
+        k8sconfig, err = k8s.cluster_config(cfg.kubeconfig, cfg.kube_ctx)
+        assert not err
 
         # Create the missing resources.
         for data_c in plan.create:
             print(f"Creating {data_c.meta.kind.upper()} "
                   f"{data_c.meta.namespace}/{data_c.meta.name}")
-            _, err = k8s.post(k8s_client, data_c.url, data_c.manifest)
+            _, err = k8s.post(k8sconfig.client, data_c.url, data_c.manifest)
             assert not err
 
         # Patch the server resources.
@@ -401,43 +445,24 @@ def apply_plan(
         for meta, patch in patches:
             print(f"Patching {meta.kind.upper()} "
                   f"{meta.namespace}/{meta.name}")
-            _, err = k8s.patch(k8s_client, patch.url, patch.ops)
+            _, err = k8s.patch(k8sconfig.client, patch.url, patch.ops)
             assert not err
 
         # Delete the excess resources.
         for data_d in plan.delete:
             print(f"Deleting {data_d.meta.kind.upper()} "
                   f"{data_d.meta.namespace}/{data_d.meta.name}")
-            _, err = k8s.delete(k8s_client, data_d.url, data_d.manifest)
+            _, err = k8s.delete(k8sconfig.client, data_d.url, data_d.manifest)
             assert not err
     except AssertionError:
-        return (None, True)
+        return True
 
     # All good.
-    return (None, False)
+    return False
 
 
-def make_plan(
-        kubeconfig: Filepath,
-        kube_ctx: Optional[str],
-        folder: Filepath,
-        selectors: Selectors,
-) -> Tuple[Optional[DeploymentPlan], bool]:
-    """Print the diff between `local_manifests` and `server_manifests`.
-
-    The diff shows what would have to change on the K8s server in order for it
-    to match the setup defined in `local_manifests`.
-
-    Inputs:
-        kubeconfig: Filepath
-            Path to Kubernetes credentials.
-        kubectx: str
-            Kubernetes context (use `None` to use the default).
-        client: `requests` session with correct K8s certificates.
-        folder: Path
-            Path to local manifests eg "./foo"
-        selectors: Selectors
-            Only operate on resources that match the selectors.
+def make_plan(cfg: Config) -> Tuple[DeploymentPlan, bool]:
+    """Return the deployment plan.
 
     Returns:
         Deployment plan.
@@ -445,130 +470,66 @@ def make_plan(
     """
     try:
         # Create properly configured Requests session to talk to K8s API.
-        (k8s_config, k8s_client), err = cluster_config(kubeconfig, kube_ctx)
-        assert not err and k8s_config and k8s_client
+        k8sconfig, err = k8s.cluster_config(cfg.kubeconfig, cfg.kube_ctx)
+        assert not err
 
         # Load manifests from local files.
-        local_meta, _, err = manio.load(folder, selectors)
-        assert not err and local_meta is not None
+        local, _, err = manio.load(cfg.folder, cfg.selectors)
+        assert not err
 
         # Download manifests from K8s.
-        server, err = manio.download(k8s_config, k8s_client, selectors)
-        assert not err and server is not None
+        server, err = manio.download(cfg, k8sconfig)
+        assert not err
 
         # Align non-plannable fields, like the ServiceAccount tokens.
-        local_meta, err = manio.align_serviceaccount(local_meta, server)
+        local_meta, err = manio.align_serviceaccount(local, server)
         assert not err
 
         # Create deployment plan.
-        plan, err = compile_plan(k8s_config, local_meta, server)
+        plan, err = compile_plan(cfg, k8sconfig, local_meta, server)
         assert not err and plan
     except AssertionError:
-        return (None, True)
+        return (DeploymentPlan(tuple(), tuple(), tuple()), True)
 
     # Print the plan and return.
     return (plan, False)
 
 
-def get_resources(
-        kubeconfig: Filepath,
-        kube_ctx: Optional[str],
-        folder: Filepath,
-        selectors: Selectors,
-        groupby: GroupBy,
-) -> Tuple[None, bool]:
-    """Download all K8s manifests and merge them into local files.
-
-    Inputs:
-        kubeconfig: Filepath
-            Path to Kubernetes credentials.
-        kubectx: str
-            Kubernetes context (use `None` to use the default).
-        folder: Path
-            Path to local manifests eg "./foo"
-        selectors: Selectors
-            Only operate on resources that match the selectors.
-        groupby: GroupBy
-            Specify relationship between new manifests and file names.
-
-    Returns:
-        None
-
-    """
-    # Use a wildcard Selector to ensure `manio.load` will read _all_ local
-    # manifests. This will allow `manio.sync` to modify the ones specified by
-    # the `selector` argument only, delete all the local manifests, and then
-    # write the new ones. This logic will ensure we never have stale manifests
-    # (see `manio.save_files` for details and how `manio.save`, which we call
-    # at the end of this function, uses it).
-    load_selectors = Selectors(kinds=SUPPORTED_KINDS, labels=None, namespaces=None)
-
+def get_resources(cfg: Config) -> bool:
+    """Download all K8s manifests and merge them into local files."""
     try:
         # Create properly configured Requests session to talk to K8s API.
-        (k8s_config, k8s_client), err = cluster_config(kubeconfig, kube_ctx)
-        assert not err and k8s_config and k8s_client
+        k8sconfig, err = k8s.cluster_config(cfg.kubeconfig, cfg.kube_ctx)
+        assert not err
+
+        # Use a wildcard Selector to ensure `manio.load` will read _all_ local
+        # manifests. This will allow `manio.sync` to modify the ones specified by
+        # the `selector` argument only, delete all the local manifests, and then
+        # write the new ones. This logic will ensure we never have stale manifests
+        # (see `manio.save_files` for details and how `manio.save`, which we call
+        # at the end of this function, uses it).
+        load_selectors = Selectors(kinds=k8sconfig.kinds,
+                                   labels=None,
+                                   namespaces=None)
 
         # Load manifests from local files.
-        _, local_files, err = manio.load(folder, load_selectors)
-        assert not err and local_files is not None
+        _, local, err = manio.load(cfg.folder, load_selectors)
+        assert not err
 
         # Download manifests from K8s.
-        server, err = manio.download(k8s_config, k8s_client, selectors)
-        assert not err and server is not None
+        server, err = manio.download(cfg, k8sconfig)
+        assert not err
 
         # Sync the server manifests into the local manifests. All this happens in
         # memory and no files will be modified here - see `manio.save` in the next step.
-        synced_manifests, err = manio.sync(local_files, server, selectors, groupby)
+        synced_manifests, err = manio.sync(local, server, cfg.selectors, cfg.groupby)
         assert not err and synced_manifests
 
         # Write the new manifest files.
-        _, err = manio.save(folder, synced_manifests)
+        err = manio.save(cfg.folder, synced_manifests, cfg.priorities)
         assert not err
     except AssertionError:
-        return (None, True)
+        return True
 
     # Success.
-    return (None, False)
-
-
-def cluster_config(
-        kubeconfig: Filepath,
-        context: Optional[str]) -> Tuple[Tuple[Optional[K8sConfig], Any], bool]:
-    """Return web session to K8s API.
-
-    This will read the Kubernetes credentials, contact Kubernetes to
-    interrogate its version and then return the configuration and web-session.
-
-    Inputs:
-        kubeconfig: str
-            Path to kubeconfig file.
-        kube_context: str
-            Kubernetes context to use (can be `None` to use default).
-
-    Returns:
-        Config, client
-
-    """
-    # Read Kubeconfig file and use it to create a `requests` client session.
-    # That session will have the proper security certificates and headers so
-    # that subsequent calls to K8s need not deal with it anymore.
-    kubeconfig = kubeconfig.expanduser()
-    try:
-        # Parse Kubeconfig file.
-        config = k8s.load_auto_config(kubeconfig, context, disable_warnings=True)
-        assert config
-
-        # Configure web session.
-        client = k8s.session(config)
-        assert client
-
-        # Contact the K8s API to update version field in `config`.
-        config, err = k8s.version(config, client)
-        assert not err and config
-    except AssertionError:
-        return ((None, None), True)
-
-    # Log the K8s API address and version.
-    logit.info(f"Kubernetes server at {config.url}")
-    logit.info(f"Kubernetes version is {config.version}")
-    return (config, client), False
+    return False
